/*  This file is part of LingoTeach, the Language Teaching program 
 *  Copyright (C) 2001-2003 The LingoTeach Team
 *
 *  This program is free software; you can redistribute it and/or modify 
 *  it under the terms of the GNU General Public License as published by 
 *  the Free Software Foundation; either version 2 of the License, or 
 *  (at your option) any later version. 
 *
 *  This program is distributed in the hope that it will be useful, 
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
 *  GNU General Public License for more details. 
 * 
 *  You should have received a copy of the GNU General Public License 
 *  along with this program; if not, write to the Free Software 
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 */

#include <gmodule.h>
#include <glib.h>
#include <string.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "lingoteach-i18n.h"
#include "module.h"
#include "errors.h"

/* module struct */
typedef struct 
{
  gchar*     (*name) ();        /* pointer to the module function mod_name */
  gchar*     (*description) (); /* pointer to the module function mod_desc */
  GtkWidget* (*box) ();         /* pointer to the module function mod_box  */
  gpointer*  (*free) ();        /* pointer to the module function mod_free */ 
  GModule*   module;            /* the module itself */
  gchar*     path;              /* the path of the module */
} lingMod;

GSList* plugins = NULL;  /* module list */

/* pototypes */
gint find_module (gpointer* mod, gpointer* name);


/*********************
 * private functions *
 *********************/

/*
 * searches a module in the list
 */
gint
find_module (gpointer* mod,  gpointer* name)
{
  lingMod* module      = (lingMod *) mod;
  gchar*   search_name = (gchar *) name;

  if (g_strncasecmp (module->name (), search_name, strlen (search_name)) == 0)
    return 0;
  return -1;
}


/********************
 * public functions *
 ********************/

/*
 * load a dynamic module
 */
void
mod_load (GtkListStore* list, char* filename)
{
  lingMod*    module;
  gchar*      (*mod_name) ();
  gchar*      (*mod_desc) ();
  GtkWidget*  (*mod_box)  ();
  gpointer*   (*mod_free) ();
  gchar*      name;
  GtkTreeIter module_iter;  

  /* allocate memory for module */
  module = g_new (lingMod, 1);
  if (module == NULL)
    {
      err_error_cb (_("Could not allocate enough memory for module!"));
      return;
    }
  
  /* open module */
  if ((module->module = g_module_open (filename, 0)) == NULL)
    {
      err_error_cb (_("Could not open module!"));
      g_free (module);
      return;
    }

  /* get the necessary module functions */
  if (!g_module_symbol (module->module, "module_name", (gpointer *) &mod_name)
      || !g_module_symbol (module->module, "module_desc", (gpointer *) &mod_desc)
      || !g_module_symbol (module->module, "module_box", (gpointer *) &mod_box)
      || !g_module_symbol (module->module, "module_free", (gpointer *) &mod_free))
    {
      err_error_cb (_("Could not find module name, description,\n"
		      " box or free method in module!"));
      g_free (module);
      return;
    }
 
  /* set the name and description */
  /* FIXME: fix compiler warning */
  module->name        = mod_name;
  module->description = mod_desc;
  module->box         = mod_box;
  module->free        = mod_free;

  /* check, if it is already loaded */
  name = module->name ();
  if (g_slist_find_custom (plugins, name, (GCompareFunc) find_module) != NULL)
    {
      err_error_cb (_("This module is already loaded."));
      g_module_close (module->module);
      g_free (module);
      return;
    }

  /* add module to the internal list */
  plugins = g_slist_append (plugins, module);

  /* add the module to the tree list */
  gtk_list_store_insert_before (list, &module_iter, NULL);
  gtk_list_store_set (list, 
		      &module_iter,
		      0, module->name (),
		      -1);

  /* save file path */
  module->path = (gchar *) malloc (strlen (filename));
  strcpy (module->path, filename);

  return;
}

/*
 * unloads (removes) a dynamic module
 */
lingbool 
mod_unload (gchar* name)
{
  lingMod* module;

  GSList* unload = g_slist_find_custom (plugins, name, 
					(GCompareFunc) find_module);
  if (unload == NULL)
    return FALSE;

  module = unload->data;
  module->free ();                 /* call the freeing method in the module */
  free (module->path);
  g_module_close (module->module);
  plugins = g_slist_remove_link (plugins, unload);
  g_slist_free_1 (unload);
  return TRUE;
}

/*
 * gets the gtk stuff from the module and applies it to the box
 */
GtkWidget*
mod_set_content (gchar* name)
{
  GSList*  active = g_slist_find_custom (plugins, name, 
					 (GCompareFunc) find_module);
  lingMod* module = active->data;
  return module->box ();
}

void
mod_get_list (GtkListStore* list)
{
  GSList*     p = plugins;
  lingMod*    module;
  GtkTreeIter module_iter;

  while (p != NULL)
    {
      module = p->data;
      
      /* add the module to the tree list */
      gtk_list_store_insert_before (list, &module_iter, NULL);
      
      gtk_list_store_set (list, 
			  &module_iter,
			  0, module->name (),
			  -1);
      p = g_slist_next (p);
    }
  return;
}

/* 
 * return the path of a module with the given name 
 */
gchar*
mod_get_path (gchar* name)
{
  GSList*  active = g_slist_find_custom (plugins, name, 
					 (GCompareFunc) find_module);
  lingMod* module = active->data;
  return module->path;
}
